from importlib.metadata import PackageNotFoundError, version

# add version information
__version__: str
"""Version info."""


def _get_version() -> str:
    """Get the current installed version of the package.

    Returns:
        str: "_unknown_" if the version metadata is missing (e.g. running from sources)
    """
    try:
        return version("colorevo")
    except PackageNotFoundError:
        # swtchd is not installed
        return "_unknown_"


__version__ = _get_version()



#!/usr/bin/env python

# This file is part of colorevo, but was adapted from the Taurus project
#
# http://taurus-scada.org
#
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus. If not, see <http://www.gnu.org/licenses/>.
#

"""This module provides the set of base classes designed to provide
configuration features to the classes that inherit from them"""

__all__ = ["configurableProperty", "BaseConfigurableClass"]

__docformat__ = "restructuredtext"


import logging


class configurableProperty(object):
    """A dummy class used to handle properties with the configuration API

    .. warning:: this class is intended for internal use by the configuration
                 package. Do not instantiate it directly in your code.
                 Use :meth:`BaseConfigurableClass.registerConfigProperty`
                  instead.
    """

    def __init__(self, name, fget, fset, obj=None):
        self.name = name
        self.fget = fget  # this may either be a method or a method name
        self.fset = fset  # this may either be a method or a method name
        self._obj = obj  # obj is only needed if fset or fget are method names

    def createConfig(self, allowUnpickable=False):
        """
        returns value returned by the fget function of this property.
        the allowUnpickable parameter is ignored
        """
        if isinstance(self.fget, str):  # fget is not a method but a method name...
            result = getattr(self._obj, self.fget)()
        else:
            result = self.fget()
        return result

    def applyConfig(self, value, depth=-1):
        """
        calls the fset function for this property with the given value.
        The depth parameter is ignored
        """
        if isinstance(self.fget, str):  # fget is not a method but a method name...
            getattr(self._obj, self.fset)(value)
        else:
            self.fset(value)

    def objectName(self):
        """returns the name of this property"""
        return self.name


class BaseConfigurableClass(object):
    """
    A base class defining the API for configurable objects.

    .. note:: One implicit requisite is that a configurable object must also
              provide a `meth:`objectName` method which returns the object
              name.
              This is typically fulfilled by inheriting from QObject.

    Using objects that inherit from :class:`BaseConfigurableClass` automates
    saving and restoring of application settings.

    The basic idea is that each object/widget in your application is
    responsible for providing a dictionary containing information on its
    properties (see :meth:`createConfig`). The same object/widget is also
    responsible for restoring such properties when provided with a
    configuration dictionary (see :meth:`applyConfig`).

    For a certain property to be saved/restored it is usually enough to
    *register* it using :meth:`registerConfigProperty`. When the objects are
    structured in a hierarchical way (e.g. as the widgets in a Qt application),
    the parent widget can (should) delegate the save/restore of its children to
    the children themselves. This delegation is done by registering the
    children using :meth:`registerConfigDelegate`.

    Consider the following example: I am creating a groupbox container which
    contains a :class:`TaurusForm` and I want to save/restore the state of the
    checkbox and the properties of the form::

        #The class looks like this:
        class MyBox(Qt.QGroupBox, BaseConfigurableClass):
            def __init__(self):
                ...
                self.form = TaurusForm()
                ...
                self.registerConfigProperty(self.isChecked, self.setChecked,
                                            'checked')
                # the TaurusForm already handles its own configuration!
                self.registerConfigDelegate(self.form)
                ...

        #and we can retrieve the configuration doing:
        b1 = MyBox()
        # we get the configuration as a dictionary
        cfg = b1.createConfig()
        # checked is a registered property of MyBox class
        b1.setChecked(True)
        # modifiableByUser is a registered property of a TaurusForm
        b1.form.setModifiableByUser(True)
        ...
        b2 = MyBox()
        #now b2 has the same configuration as b1 when cfg was created
        b2.applyConfig(cfg)

    :meth:`createConfig` and :meth:`applyConfig` methods use a dictionary for
    passing the configuration, but :class:`BaseConfigurableClass` also provides
    some other convenience methods for working with files
    (:meth:`saveConfigFile` and :meth:`loadConfigFile`) or as QByteArrays
    (:meth:`createQConfig` and :meth:`applyQConfig`)

    Finally, we recommend to use :class:`TaurusMainWindow` for all Taurus GUIs
    since it automates all the steps for *saving properties when closing* and
    *restoring the settings on startup*. It also provides a mechanism for
    implementing "perspectives" in your application.

    """

    defaultConfigRecursionDepth = -1
    # the latest element of this list is considered the current version
    _supportedConfigVersions = ("__UNVERSIONED__",)

    def __init__(self, *args, **kwargs):
        self.resetConfigurableItems()

    @staticmethod
    def isConfigDict(x):
        """Checks if the given argument has the structure of a configdict

        :param x: (object) object to test

        :return: (bool) True if it is a configdict, False otherwise.
        """
        if not isinstance(x, dict):
            return False
        for k in (
            "__orderedConfigNames__",
            "__itemConfigurations__",
            "ConfigVersion",
            "__pickable__",
        ):
            if k not in x:
                return False

        for k in x["__orderedConfigNames__"]:
            if k not in x["__itemConfigurations__"]:
                logging.warning('missing configuration for "%s" in %r', k, x)
        return True

    def createConfig(self, allowUnpickable=False):
        """
        Returns a dictionary containing configuration information about the
        current state of the object.

        In most usual situations, using :meth:`registerConfigProperty` and
        :meth:`registerConfigDelegate`, should be enough to cover all needs using
        this method, although it can be reimplemented in children classes to support
        very specific configurations.

        By default, meth:`createQConfig` and meth:`saveConfigFile` call to this
        method for obtaining the data.

        Hint: The following code allows you to serialize the configuration
        dictionary as a string (which you can store as a QSetting, or as a Tango
        Attribute, provided that allowUnpickable==False)::

            import pickle
            s = pickle.dumps(widget.createConfig())  #s is a string that can be stored

        :param alllowUnpickable:  (bool) if False the returned dict is
                                  guaranteed to be a pickable object. This is
                                  the default and preferred option because it
                                  allows the serialization as a string that can
                                  be directly stored in a QSetting. If True, this
                                  limitation is not enforced, which allows to
                                  use more complex objects as values (but limits
                                  its persistence).

        :return: (dict<str,object>) configurations (which can be loaded with :meth:`applyConfig`).

        .. seealso: :meth:`applyConfig` , :meth:`registerConfigurableItem`,  meth:`createQConfig`, meth:`saveConfigFile`
        """
        configdict = {
            "ConfigVersion": self._supportedConfigVersions[-1],
            "__pickable__": True,
        }
        # store the configurations for all registered configurable items as
        # well
        itemcfgs = {}
        for k, v in self.__configurableItems.items():
            itemcfgs[k] = v.createConfig(allowUnpickable=allowUnpickable)
        configdict["__itemConfigurations__"] = itemcfgs
        configdict["__orderedConfigNames__"] = self.__configurableItemNames
        return configdict

    def applyConfig(self, configdict, depth=None):
        """
        applies the settings stored in a configdict to the current object.

        In most usual situations, using :meth:`registerConfigProperty` and
        :meth:`registerConfigDelegate`, should be enough to cover all needs
        using this method, although it can be reimplemented in children classes
        to support very specific configurations.

        :param configdict: (dict)
        :param depth: (int)  If depth = 0, applyConfig will only be called
                      for this object, and not for any other object registered
                      via :meth:`registerConfigurableItem`. If depth > 0,
                      applyConfig will be called recursively as many times as
                      the depth value. If depth < 0 (default, see note), no
                      limit is imposed to recursion (i.e., it will recurse for
                      as deep as there are registered items).

        .. note:: the default recursion depth can be tweaked in derived classes
                  by changing the class property `defaultConfigRecursionDepth`

        .. seealso:: :meth:`createConfig`
        """
        if depth is None:
            depth = self.defaultConfigRecursionDepth
        if not self.checkConfigVersion(configdict):
            raise ValueError("the given configuration is of unsupported version")
        # delegate restoring the configuration of any registered configurable
        # item
        if depth != 0:
            itemcfgs = configdict["__itemConfigurations__"]
            # we use the sorted item names that was stored in the configdict
            for key in configdict["__orderedConfigNames__"]:
                if key in self.__configurableItems:
                    self.__configurableItems[key].applyConfig(
                        itemcfgs[key], depth=depth - 1
                    )

    def getConfigurableItemNames(self):
        """
        returns an ordered list of the names of currently registered
        configuration items (delegates and properties)

        :return: (list<unicode>)
        """
        return self.__configurableItemNames

    def resetConfigurableItems(self):
        """
        clears the record of configurable items depending of this object

        .. seealso:: :meth:`registerConfigurableItem`
        """
        self.__configurableItemNames = []
        self.__configurableItems = {}

    def registerConfigDelegate(self, delegate, name=None):
        """
        Registers the given object as a delegate for configuration.
        Delegates are typically other objects inheriting from
        BaseConfigurableClass
        (or at least they must provide the following methods:

          - `createConfig` (as provided by, e.g., BaseConfigurableClass)
          - `applyConfig` (as provided by, e.g., BaseConfigurableClass)
          - `objectName` (as provided by, e.g., QObject)

        :param delegate: (BaseConfigurableClass) The delegate object to be
                         registered.
        :param name: (str) The name to be used as a key for this item in the
                     configuration dictionary. If None given, the object name
                     is used by default.

        .. note:: the registration order will be used when restoring
                  configurations

        .. seealso:: :meth:`unregisterConfigurableItem`,
                     :meth:`registerConfigProperty`, :meth:`createConfig`
        """
        return self._registerConfigurableItem(delegate, name=name)

    def registerConfigProperty(self, fget, fset, name):
        """
        Registers a certain property to be included in the config dictionary.

        In this context a "property" is some named value that can be obtained
        via a getter method and can be set via a setter method.

        :param fget: (method or str) method (or name of a method) that gets no
                     arguments and returns the value of a property.
        :param fset: (method or str) method (or name of a method) that gets as
                     an argument the value of a property, and sets it
        :param name: (str) The name to be used as a key for this property in
                     the configuration dictionary

        .. note:: the registration order will be used when restoring
                  configurations

        .. seealso:: :meth:`unregisterConfigurableItem`,
                     :meth:`registerConfigDelegate`,
                     :meth:`createConfig`
        """
        if isinstance(fget, str) or isinstance(fset, str):
            import weakref

            obj = weakref.proxy(self)
        else:
            obj = None
        p = configurableProperty(name, fget, fset, obj=obj)
        return self._registerConfigurableItem(p, name=name)

    def _registerConfigurableItem(self, item, name=None):
        """
        Registers the given item as a configurable item which depends of this
        class.

        .. note:: This method is not meant to be called directly. Use
                  :meth:`registerConfigProperty`,
                  :meth:`registerConfigDelegate`
                  instead

        Registered items are expected to implement the
        following methods:
          - `createConfig` (as provided by, e.g., BaseConfigurableClass)
          - `applyConfig` (as provided by, e.g., BaseConfigurableClass)
          - `objectName` (as provided by, e.g., QObject)

        :param item: (object) The object that should be registered.
        :param name: (str) The name to be used as a key for this item in the
                     configuration dictionary. If None given, the object name
                     is used by default.

        .. note:: the registration order will be used when restoring
                  configurations

        .. seealso:: :meth:`unregisterConfigurableItem`, :meth:`createConfig`
        """
        if name is None:
            name = item.objectName()
        name = str(name)
        if name in self.__configurableItemNames:
            # abort if duplicated names
            raise ValueError(
                (
                    '_registerConfigurableItem: An object with name "%s" is '
                    + "already registered"
                )
                % name
            )
        self.__configurableItemNames.append(name)
        self.__configurableItems[name] = item

    def unregisterConfigurableItem(self, item, raiseOnError=True):
        """
        unregisters the given item (either a delegate or a property) from the
        configurable items record. It raises an exception if the item is not
        registered

        :param item: (object or str) The object that should be unregistered.
                     Alternatively, the name under which the object was
                     registered can be passed as a python string.
        :param raiseOnError: (bool) If True (default), it raises a KeyError
                             exception if item was not registered. If False, it
                             just logs a debug message

        .. seealso:: :meth:`registerConfigProperty`,
                     :meth:`registerConfigDelegate`
        """
        if isinstance(item, str):
            name = item
        else:
            name = str(item.objectName())
        if name in self.__configurableItemNames and name in self.__configurableItems:
            self.__configurableItemNames.remove(name)
            self.__configurableItems.pop(name)
            return True
        elif raiseOnError:
            raise KeyError('"%s" was not registered.' % name)
        else:
            logging.debug('"%s" was not registered. Skipping' % name)
            return False

    def checkConfigVersion(self, configdict, showDialog=False, supportedVersions=None):
        """
        Check if the version of configdict is supported. By default, the
        BaseConfigurableClass objects have ["__UNVERSIONED__"] as their list
        of supported versions, so unversioned config dicts will be accepted.

        :param configdict: (dict) configuration dictionary to check
        :param showDialog: (bool) whether to show a QtWarning dialog if check
                           failed (false by default)
        :param supportedVersions: (sequence<str>, or None) supported version
                                  numbers, if None given, the versions
                                  supported by this widget will be used
                                  (i.e., those defined in
                                  self._supportedConfigVersions)

        :return: (bool) returns True if the configdict is of the right version
        """
        if supportedVersions is None:
            supportedVersions = self._supportedConfigVersions
        version = configdict.get("ConfigVersion", "__UNVERSIONED__")
        if version not in supportedVersions:
            msg = "Unsupported Config Version %s. (Supported: %s)" % (
                version,
                repr(supportedVersions),
            )
            logging.warning(msg)
            return False
        return True

    def createQConfig(self):
        """
        returns the current configuration status encoded as a QByteArray. This
        state can therefore be easily stored using QSettings

        :return: (QByteArray) (in the current implementation this is just a
                 pickled configdict encoded as a QByteArray

        .. seealso:: :meth:`restoreQConfig`
        """
        from PyQt6 import QtCore
        import pickle

        configdict = self.createConfig(allowUnpickable=False)
        return QtCore.QByteArray(pickle.dumps(configdict))

    def applyQConfig(self, qstate):
        """
        restores the configuration from a qstate generated by :meth:`getQState`

        :param qstate: (QByteArray)

        .. seealso:: :meth:`createQConfig`
        """
        if qstate.isNull():
            return
        import pickle

        configdict = pickle.loads(qstate.data())
        self.applyConfig(configdict)

    def saveConfigFile(self, fname):
        """Stores the current configuration on a file

        :param ofile: (file) filename  or file descriptorto store the
                      configuration

        :return: (str) file name used
        """
        import pickle

        configdict = self.createConfig(allowUnpickable=False)
        logging.info("Saving current settings in '%s'" % fname)

        with open(fname, "wb") as f:
            pickle.dump(configdict, f)
        return fname

    def loadConfigFile(self, fname):
        """Reads a file stored by :meth:`saveConfig` and applies the settings

        :param ifile: (file or string) file or filename from where to read the
                      configuration

        :return: (str) file name used
        """
        import pickle

        with open(fname, "rb") as f:
            configdict = pickle.load(f)
        self.applyConfig(configdict)
        return fname



#  Copyright 2017 Carlos Pascual-Izarra <cpascual@users.sourceforge.net>
#
#  This file is part of colorevo.
#
#  PeakEvo is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  PeakEvo is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with colorevo. If not, see <http://www.gnu.org/licenses/>.


import cv2
import time
from .h5writer import H5Writer


class DummyWriter:
    """
    Dummy writer that does not write anything
    """

    def start(self, *args, **kwargs):
        pass

    def write_frame(self, *args, **kwargs):
        pass

    def end(self, *args, **kwargs):
        pass


class AcqLoop:
    """
    Acquisition loop that captures images with opencv,
    saves them into an h5file file and makes them available
    as a shared numpy mem map
    """

    def __init__(self, src):
        self.cap = cv2.VideoCapture(src)
        self.data = None
        self._acquiring = False
        self.t_offset = None
        self.period = None
        self.writer = DummyWriter()
        self.reset(None)

    def reset(self, outfile, max_fps=None):
        # stop writing
        self.writer.end()
        self.writer = DummyWriter()

        if max_fps is None:
            self.period = None
        else:
            self.period = 1.0 / max_fps

        # capture one frame to initialize the writer
        ok, frame = self.cap.read()
        if not ok:
            raise RuntimeError("problem capturing from source")

        # store the absolute start time
        self.t_offset = time.time()
        self.data = frame, 0
        if outfile is not None:
            self.writer = H5Writer(outfile, frame.shape, frame.dtype)
            self.writer.start()

    def run(self):
        self._acquiring = True
        while self._acquiring:
            t0 = time.time()
            _, frame = self.cap.read()
            t = time.time() - self.t_offset
            self.data = frame, t
            self.writer.write_frame(frame, t)
            # print('saved', t, self.writer)
            if self.period is not None:
                rest = self.period - (time.time() - t0)
                if rest >= 0:
                    time.sleep(rest)
                else:
                    print(
                        "Cannot reach {}fps. Achieved:{:.0f}fps".format(
                            1 / self.period, 1 / (self.period - rest)
                        )
                    )

    def stop(self):
        self._acquiring = False
        self.writer.end()


if __name__ == "__main__":
    import os
    import sys
    import tempfile
    import threading
    from PyQt6.QtCore import QTimer, QCoreApplication

    class Reader:
        def __init__(self):
            super(Reader, self).__init__()

            self.timer = QTimer()
            self.acq_loop = None
            self.timer.timeout.connect(self.process)
            self._last_processed = -1

        def start(self):
            # create the acq loop
            ofile = os.path.join(tempfile.gettempdir(), "kkcap.h5")
            self.acq_loop = AcqLoop(0, ofile)

            # create the acq thread and start it
            self.thread = threading.Thread(target=self.acq_loop.run, daemon=True)
            self.thread.start()
            self.timer.start(200)

        def process(self):
            try:
                fr, t = self.acq_loop.data
            except Exception:
                return
            if t <= self._last_processed:
                print("no new data. skipping")
                return
            self._last_processed = t
            print("processing... ", self._last_processed, fr.shape, fr.mean())

            if self._last_processed > 3:
                self.stop()

            time.sleep(0.5)

        def stop(self):
            print("STOPPING")
            self.acq_loop.stop()
            self.timer.stop()

    app = QCoreApplication(sys.argv)

    r = Reader()
    r.start()

    sys.exit(app.exec())



#  Copyright 2019 Carlos Pascual-Izarra <cpascual@users.sourceforge.net>
#
#  This file is part of colorevo.
#
#  colorevo is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  colorevo is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import itertools


def bgr2h(arr):
    """Obtain hue from BGR image.
    Code adapted from matplotlib.colors.rgb_to_hsv

    See: https://en.wikipedia.org/wiki/HSL_and_HSV#Hue_and_chroma
    """
    arr = arr.astype(np.promote_types(arr.dtype, np.float32))
    b, g, r = 0, 1, 2
    out = np.zeros_like(arr[..., b])
    arr_max = arr.max(-1)
    ipos = arr_max > 0
    delta = arr.ptp(-1)
    # Saturation (s) is not needed. Commented out for efficiency
    # s = np.zeros_like(delta)
    # s[ipos] = delta[ipos] / arr_max[ipos]
    ipos = delta > 0
    # red is max
    idx = (arr[..., r] == arr_max) & ipos
    out[idx] = (arr[idx, g] - arr[idx, b]) / delta[idx]
    # green is max
    idx = (arr[..., g] == arr_max) & ipos
    out[idx] = 2.0 + (arr[idx, b] - arr[idx, r]) / delta[idx]
    # blue is max
    idx = (arr[..., b] == arr_max) & ipos
    out[idx] = 4.0 + (arr[idx, r] - arr[idx, g]) / delta[idx]

    out = (out / 6.0) % 1.0

    return out


def bgr2hsv(arr):
    """
    Convert an array of float BGR values (in the range [0, 1]) to HSV values.

    Code adapted from matplotlib.colors.rgb_to_hsv

    See: https://en.wikipedia.org/wiki/HSL_and_HSV#Hue_and_chroma

    Parameters
    ----------
    arr : (..., 3) array-like
       All values must be in the range [0, 1]

    Returns
    -------
    (..., 3) `~numpy.ndarray`
       Colors converted to HSV values in range [0, 1]
    """
    arr = np.asarray(arr)
    b, g, r = 0, 1, 2

    # check length of the last dimension, should be _some_ sort of rgb
    if arr.shape[-1] != 3:
        raise ValueError(
            "Last dimension of input array must be 3; " f"shape {arr.shape} was found."
        )

    in_shape = arr.shape
    arr = np.array(
        arr,
        dtype=np.promote_types(arr.dtype, np.float32),  # Don't work on ints.
        ndmin=2,  # In case input was 1D.
    )
    out = np.zeros_like(arr)
    arr_max = arr.max(-1)
    ipos = arr_max > 0
    delta = np.ptp(arr, -1)
    s = np.zeros_like(delta)
    s[ipos] = delta[ipos] / arr_max[ipos]
    ipos = delta > 0
    # red is max
    idx = (arr[..., r] == arr_max) & ipos
    out[idx, 0] = (arr[idx, g] - arr[idx, b]) / delta[idx]
    # green is max
    idx = (arr[..., g] == arr_max) & ipos
    out[idx, 0] = 2.0 + (arr[idx, b] - arr[idx, r]) / delta[idx]
    # blue is max
    idx = (arr[..., b] == arr_max) & ipos
    out[idx, 0] = 4.0 + (arr[idx, r] - arr[idx, g]) / delta[idx]

    out[..., 0] = (out[..., 0] / 6.0) % 1.0
    out[..., 1] = s
    out[..., 2] = arr_max

    return out.reshape(in_shape)


def bgr2rgb(arr):
    return arr[..., ::-1]


def hsv_generator(h0=0, hn=15, s=(0.99, 0.5), v=0.99):
    """
    Generator that returns evenly-spaced non-repeated hsv colors.
    It varies the H using the golden ratio, and cycles the saturation
    according to the given values.
    Inspired in:
     https://martin.ankerl.com/2009/12/09/how-to-create-random-colors-programmatically/
    """
    gr = 0.618033988749895  # golden_ratio_conjugate
    h = h0
    s_cycle = itertools.cycle(s)
    while True:
        s = next(s_cycle)
        for _ in range(hn):
            yield h, s, v
            h = (h + gr) % 1


if __name__ == "__main__":
    # from matplotlib.colors import rgb_to_hsv
    #
    # a = np.random.rand(100, 200, 3)
    # print(np.allclose(bgr2h(a), rgb_to_hsv(a[..., ::-1])[...,0]))
    #
    # a = np.arange(100*200*3).reshape(100, 200, 3)
    # print(np.allclose(bgr2h(a), rgb_to_hsv(a[..., ::-1])[...,0]))

    colors = hsv_generator()
    for i in range(60):
        print(i, next(colors))



#  Copyright 2019 Carlos Pascual-Izarra <cpascual@users.sourceforge.net>
#
#  This file is part of colorevo.
#
#  colorevo is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  colorevo is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.


from datetime import datetime
import h5py


class H5Writer:
    """
    Saves frames and time to a hdf5 file file.
    Shamelessly adapted (simplified) from sardana.macroserver.recorders.h5storage
    (https://github.com/sardana-org/sardana )
    """

    def __init__(self, filename, frame_shape, frame_dtype):
        self.fd = fd = h5py.File(filename, mode="w")
        self._write_enabled = False

        fd.attrs["creator"] = f"colorevo's {self.__class__.__name__}"
        fd.attrs["HDF5_Version"] = h5py.version.hdf5_version
        fd.attrs["h5py_version"] = h5py.version.version

        self._frames = fd.create_dataset(
            "frames",
            dtype=frame_dtype,
            shape=([0] + list(frame_shape)),
            maxshape=([None] + list(frame_shape)),
            chunks=(1,) + tuple(frame_shape),
            compression="gzip",
        )

        self._times = fd.create_dataset(
            "times", dtype="float", shape=(0,), maxshape=(None,)
        )

        self._fr_count = 0

        fd.flush()

    def start(self):
        self.fd.attrs["start_time"] = datetime.now().isoformat()
        self._write_enabled = True

    def write_frame(self, frame, fr_time):
        if not self._write_enabled:
            return
        self._fr_count += 1
        self._frames.resize(self._fr_count, axis=0)
        self._times.resize(self._fr_count, axis=0)

        self._frames[self._fr_count - 1, ...] = frame
        self._times[self._fr_count - 1, ...] = fr_time

        self.fd.flush()

    def end(self):
        self.fd.attrs["end_time"] = datetime.now().isoformat()
        self._write_enabled = False
        self.fd.close()


if __name__ == "__main__":
    import os
    import tempfile
    import numpy
    import time

    sh = (4, 5)
    dt = "int8"

    fname = os.path.join(tempfile.gettempdir(), "kk.h5")
    w = H5Writer(fname, sh, dt)

    w.start()

    for i in range(9):
        f = numpy.random.randint(0, 100, sh)
        t = time.time()
        # print(f,t)
        w.write_frame(f, t)

    w.end()

    print(f"//////////// {fname} /////////////")

    def pp(name):
        e = f[name]
        print(e)
        print(f"{name}.attrs: {list(e.attrs)}")

    f = h5py.File(fname)
    print(f, list(f.attrs))
    f.visit(pp)



from click.testing import CliRunner


def test_smoke_cli():
    from colorevo import __version__
    from colorevo.maingui import main

    runner = CliRunner()
    result = runner.invoke(main, args=["--version"])
    assert result.exit_code == 0
    assert __version__ in result.stdout

    result = runner.invoke(main, args=["--help"])
    assert result.exit_code == 0
    assert "--reset" in result.stdout
    assert "--keep-alive" in result.stdout
    assert "--no-keep-alive" in result.stdout



#  Copyright 2019 Carlos Pascual-Izarra <cpascual@users.sourceforge.net>
#
#  This file is part of colorevo.
#
#  colorevo is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  colorevo is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.


import sys
import os
import re
import datetime
import logging
import threading
import numpy
import shutil
import click
import h5py
import json
from tempfile import gettempdir
from PyQt6 import QtWidgets, QtCore, QtGui
import pyqtgraph as pg

from colorevo import __version__
from .configuration import BaseConfigurableClass
from .colors import bgr2hsv, bgr2rgb, hsv_generator
from .acq_cv2 import AcqLoop
from .h5writer import H5Writer

# Set logging
logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)

# Set interpretation of images as row-major (to avoid turned video)
pg.setConfigOptions(imageAxisOrder="row-major")


class ColorEvoMain(QtWidgets.QMainWindow, BaseConfigurableClass):
    itercolors = hsv_generator()
    _tmp_rec_file_name = os.path.join(gettempdir(), "colorevo_tmp_rec.h5")
    refresh_timer_period = 500  # ms
    workdir = os.path.join(os.path.dirname(__file__), os.path.pardir, "data")

    def __init__(self, parent=None, load_settings=True):
        super().__init__(parent=parent)
        self._qsettings = QtCore.QSettings()

        # initialize icon search path
        curdir = os.path.dirname(os.path.realpath(__file__))
        QtCore.QDir.addSearchPath("icons", os.path.join(curdir, "icons"))

        # timer for refreshing the plots in live operation
        self.refresh_timer = QtCore.QTimer()
        self.refresh_timer.timeout.connect(self.onRefresh)

        # initializations
        self.acq_loop = None
        self.acq_thread = None
        self._last_processed = -1
        self._h5file = None
        self._h5times = None

        self.imgPlot = pg.PlotWidget()
        self.evoPlot_h = pg.PlotWidget()
        self.evoPlot_s = pg.PlotWidget()
        self.evoPlot_v = pg.PlotWidget()
        self.toolBar = self.addToolBar("Toolbar")
        self.toolBar.setObjectName("mainTB")
        self.rois = []
        self.h_curves = []
        self.s_curves = []
        self.v_curves = []

        # allow dockwidgets to be placed anywhere
        self.setDockNestingEnabled(True)

        self._initImgPlot()
        self._initEvoPlots()
        self._initToolBar()
        self._initMenus()

        # register config properties
        self.registerConfigProperty(self.getCamera, self.setCamera, "camera")
        self.registerConfigProperty(self.getROIsConfig, self.restoreROIs, "ROIs")
        self.registerConfigProperty(self.fpsSB.value, self.fpsSB.setValue, "fps")
        self.registerConfigProperty(
            self._getTmpRecFile, self._setTmpRecFile, "tmp_rec_file"
        )

        # Load config settings
        if load_settings:
            self.loadSettings()

        # we are ready
        self.statusBar().showMessage("Started", 3000)

    def _initImgPlot(self):
        self.imgPlot.setAspectLocked(True)
        self.imgPlot.getPlotItem().getViewBox().invertY(True)

        # Add img itrm
        self.img = pg.ImageItem()
        self.imgPlot.addItem(self.img)

        # set up dockwidget
        dw = QtWidgets.QDockWidget("Image")
        dw.setObjectName("imageDW")
        dw.setWidget(self.imgPlot)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, dw)
        self._imageDW = dw

    def _initEvoPlots(self):
        # configure plot
        self.evoPlot_h.setLabels(left=("hue",), bottom=("time (s)",))
        self.evoPlot_s.setLabels(left=("saturation",), bottom=("time (s)",))
        self.evoPlot_v.setLabels(left=("brightness",), bottom=("time (s)",))

        # add time selector (the time_selector_h acts as main and the other 2 as slaves)
        self.time_selector_h = pg.InfiniteLine(pos=0, angle=90, movable=True)
        self.time_selector_s = pg.InfiniteLine(pos=0, angle=90, movable=True)
        self.time_selector_v = pg.InfiniteLine(pos=0, angle=90, movable=True)
        self.time_selector_h.sigPositionChanged.connect(self.onMainTimeSelectorChanged)
        self.time_selector_s.sigPositionChanged.connect(self.onSlaveTimeSelectorChanged)
        self.time_selector_v.sigPositionChanged.connect(self.onSlaveTimeSelectorChanged)

        # set up dockwidgets
        dw = QtWidgets.QDockWidget("Hue Values")
        dw.setObjectName("huesDW")
        dw.setWidget(self.evoPlot_h)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, dw)
        self._hueDW = dw

        dw = QtWidgets.QDockWidget("Saturation Values")
        dw.setObjectName("saturationDW")
        dw.setWidget(self.evoPlot_s)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, dw)
        self._saturationDW = dw

        dw = QtWidgets.QDockWidget("Brightness Values")
        dw.setObjectName("brightnessDW")
        dw.setWidget(self.evoPlot_v)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, dw)
        self._brightnessDW = dw

    def _initToolBar(self):
        # video file selector
        self.loadVideoAction = self.toolBar.addAction(
            QtGui.QIcon("icons:video-x-generic.svg"), "Open data source file"
        )
        self.loadVideoAction.setCheckable(True)
        self.loadVideoAction.toggled.connect(self.onLoadVideo)

        self.toolBar.addSeparator()
        # start/stop record button
        rec_icon = QtGui.QIcon()
        rec_icon.addPixmap(
            QtGui.QPixmap("icons:media-record.svg"),
            rec_icon.Mode.Normal,
            rec_icon.State.Off,
        )
        rec_icon.addPixmap(
            QtGui.QPixmap("icons:media-playback-stop.svg"),
            rec_icon.Mode.Normal,
            rec_icon.State.On,
        )
        self.recordAction = self.toolBar.addAction("Start/Stop recording")
        self.recordAction.setIcon(rec_icon)
        self.recordAction.setCheckable(True)
        self.recordAction.setChecked(False)
        self.recordAction.setEnabled(False)
        self.recordAction.toggled.connect(self.onRecordFromCameraActionToggled)

        # FPS spinbox
        self.fpsSB = sb = QtWidgets.QDoubleSpinBox()
        sb.setRange(0.1, 40)
        sb.setSuffix("fps")
        sb.setValue(24)
        sb.setSpecialValueText("stopped")
        sb.setToolTip("Maximum fps for file recording")
        self.toolBar.addWidget(sb)
        sb.valueChanged[float].connect(self.setAcqFps)

        self.toolBar.addSeparator()

        # Add rect ROI action
        self.addRectROIAction = self.toolBar.addAction(
            QtGui.QIcon("icons:draw-rectangle.svg"), "Add new rectangular ROI"
        )
        self.addRectROIAction.triggered.connect(self.addRectROI)

        # Add polyline ROI action
        self.addPolyLineROIAction = self.toolBar.addAction(
            QtGui.QIcon("icons:draw-polyline.svg"), "Add new polyline ROI"
        )
        self.addPolyLineROIAction.triggered.connect(self.addPolyLineROI)

        # Reset evoPlot action
        self.refreshEvoAction = self.toolBar.addAction(
            QtGui.QIcon("icons:view-refresh.svg"),
            "Recalculate color values (uncheck to abort)",
        )
        self.refreshEvoAction.setCheckable(True)
        self.refreshEvoAction.toggled.connect(self.refreshEvoPlot)

        # Clear all ROIs action
        self.removeAllRoisAction = self.toolBar.addAction(
            QtGui.QIcon("icons:edit-clear-history.svg"),
            "Remove all current ROIs",
        )
        self.removeAllRoisAction.triggered.connect(self.removeAllRois)

    def _initMenus(self):
        # File Menu
        file_menu = self.menuBar().addMenu("&File")

        # Save all ROIs action
        self.saveAllRoisAction = file_menu.addAction(
            QtGui.QIcon("icons:document-save-as-template.svg"),
            "Save current ROIs configuration to a file",
        )
        self.saveAllRoisAction.triggered.connect(self.onSaveAllRoisAction)

        # Restore ROIs action
        self.restoreRoisAction = file_menu.addAction(
            QtGui.QIcon("icons:document-new-from-template.svg"),
            "Restore ROI from a file",
        )
        self.restoreRoisAction.triggered.connect(self.onRestoreRoisAction)

        file_menu.addSeparator()

        # Other settings action
        self.setTmpRecDirAction = file_menu.addAction("Set temporary data file name...")
        self.setTmpRecDirAction.triggered.connect(self._setTmpRecFile)

        # Camera menu
        camera_menu = self.menuBar().addMenu("&Camera")
        self._cameraActions = QtGui.QActionGroup(self)

        camera_names = ["Disabled"]

        # add openCV cameras
        camera_names += [f"CV{id}" for id in range(8)]

        # create all camera actions
        for name in camera_names:
            a = self._cameraActions.addAction(name)
            a.setCheckable(True)

        # check the "Disabled" action by default
        self._cameraActions.actions()[0].setChecked(True)

        camera_menu.addActions(self._cameraActions.actions())

        self._cameraActions.triggered.connect(self.onCameraChanged)

        # View menu
        view_menu = self.menuBar().addMenu("&View")
        view_menu.addAction(self._imageDW.toggleViewAction())
        view_menu.addAction(self._hueDW.toggleViewAction())
        view_menu.addAction(self._saturationDW.toggleViewAction())
        view_menu.addAction(self._brightnessDW.toggleViewAction())
        view_menu.addAction(self.toolBar.toggleViewAction())

    def onLoadVideo(self, checked):
        if not checked:
            return

        FRAMES_FLT = "Frames files (*.h5 *.hdf *hdf5)"
        VIDEOS_FLT = "Videos (*.avi *.mp4 *.mpeg *.mpg *.mov *.wmv)"
        OTHERS_FLT = "Other (*)"

        fname, flt = QtWidgets.QFileDialog.getOpenFileName(
            parent=self,
            caption="Open File",
            directory=self.workdir,
            filter=";;".join((FRAMES_FLT, VIDEOS_FLT, OTHERS_FLT)),
        )
        if not fname:
            logging.info("No data/video selected.")
            self.loadVideoAction.setChecked(False)
            return

        self.setCamera("Disabled")
        if self._h5file is not None:
            self._h5file.close()

        if flt == FRAMES_FLT:
            if self._h5file is not None:
                self._h5file.close()

            self.loadH5Frames(fname)

        elif flt in (VIDEOS_FLT, OTHERS_FLT):
            self.loadCVVideo(fname)

        # uncheck the loadVideo action
        self.loadVideoAction.setChecked(False)

    def loadH5Frames(self, fname):
        """Open h5 and load times and frames"""
        if self._h5file is not None:
            self._h5file.close()
        self._h5file = h5py.File(fname)
        self._h5times = self._h5file["times"][()]  # read it all to memory
        t0 = self._h5times[0]
        tf = self._h5times[-1]
        self.evoPlot_h.setXRange(t0, tf)
        self.evoPlot_s.setXRange(t0, tf)
        self.evoPlot_v.setXRange(t0, tf)
        # show time selectors (moving time_selector_h moves the other ones too)
        self.time_selector_h.setPos(t0)  # also updates the img (connected)
        self.time_selector_h.setBounds((t0, tf))
        self.time_selector_s.setBounds((t0, tf))
        self.time_selector_v.setBounds((t0, tf))
        self.evoPlot_h.addItem(self.time_selector_h)
        self.evoPlot_s.addItem(self.time_selector_s)
        self.evoPlot_v.addItem(self.time_selector_v)

    def loadCVVideo(self, fname):
        """
        Extract frames from a video file and write them in a h5file.
        Then open the h5file.
        """
        try:
            import cv2

            cap = cv2.VideoCapture(fname)
            if not cap.isOpened():
                raise IOError("CV capture failed to open file")
        except Exception as e:
            logging.error(f"Problem opening {fname}. Reason: {e}")
            return

        h5fname, flt = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            caption=f"Save frames from {os.path.basename(fname)}",
            directory=os.path.splitext(fname)[0] + ".h5",
            filter="HDF5 (*.h5)",
        )
        if not h5fname:
            return

        # capture one frame to initialize the writer
        ok, frame = cap.read()
        t = cap.get(cv2.CAP_PROP_POS_MSEC) / 1000
        if not ok:
            logging.error(f"problem capturing from {fname}")

        app = QtWidgets.QApplication.instance()
        writer = H5Writer(h5fname, frame.shape, frame.dtype)
        writer.start()
        msg = "Extracting frames..."
        nframes = cap.get(cv2.CAP_PROP_FRAME_COUNT)
        while ok:
            if not self.loadVideoAction.isChecked():
                self.statusBar().showMessage(msg + "Aborted", 3000)
                break  # abort processing if refresh button is unchecked
            writer.write_frame(frame, t)
            ok, frame = cap.read()
            t = cap.get(cv2.CAP_PROP_POS_MSEC) / 1000
            i = cap.get(cv2.CAP_PROP_POS_FRAMES)
            self.statusBar().showMessage(msg + f"({i}/{nframes})", 3000)
            app.processEvents()
        writer.end()

        # now simply open the h5file
        self.loadH5Frames(h5fname)

    def onCameraChanged(self):
        self.setCamera(self.getCamera())

    def getCamera(self):
        action = self._cameraActions.checkedAction()
        if action is None:
            return None
        return action.text()

    def setCamera(self, camera):
        # make sure the corresponding action is checked
        for c in self._cameraActions.actions():
            if c.text() == camera:
                if not c.isChecked():
                    c.setChecked(True)
                break
        else:
            # if we land here, it means camera is not known
            logging.warning(f"Unknown camera '{camera}'")
            return

        cv = re.fullmatch(r"CV([0-9]+)", camera)

        # stop recording
        self.recordAction.setChecked(False)

        # stop current AcqLoop
        if self.acq_loop is not None:
            self.acq_loop.stop()
            self.acq_thread.join()
            self.acq_loop.cap.release()
        self.recordAction.setEnabled(False)

        if camera == "Disabled":
            logging.info("Camera disabled")
            self.refresh_timer.stop()
            self.resetPlots()
            return
        elif cv:
            src = int(cv.groups()[0])
            try:
                self.acq_loop = AcqLoop(src)
            except Exception as e:
                logging.warning(f"Problem acquiring with {camera}: {e}")
                return
            self.recordAction.setEnabled(True)
        else:
            logging.warning(f'Unsupported source of video "{camera}"')
            return

        # remove time selectors
        self.evoPlot_h.removeItem(self.time_selector_h)
        self.evoPlot_s.removeItem(self.time_selector_s)
        self.evoPlot_v.removeItem(self.time_selector_v)

        # close and forget h5file if it was open
        if self._h5file is not None:
            self._h5file.close()
        self._h5file = None
        self._h5times = None

        # start acq loop (in new thread)
        self.acq_thread = threading.Thread(
            target=self.acq_loop.run, name="acq", daemon=True
        )
        self.acq_thread.start()

        # start refreshing the plots
        self.refresh_timer.start(self.refresh_timer_period)

    def onRecordFromCameraActionToggled(self, toggled):
        if toggled:
            # reset the acq loop with the tmp file
            self.acq_loop.reset(self._tmp_rec_file_name, self.fpsSB.value())
            self._last_processed = -1
            self.resetPlots()
        else:
            # reset the acq loop with the tmp file
            self.acq_loop.reset(None)
            self._last_processed = -1
            self.resetPlots()

            now = datetime.datetime.now()
            size = os.path.getsize(self._tmp_rec_file_name) // 1_048_576  # in Mb
            msg = f"{self._tmp_rec_file_name}: {size}Mb"
            self.statusBar().showMessage(msg)
            while True:  # loop until data is saved or discard is forced
                # offer to save data
                fname, flt = QtWidgets.QFileDialog.getSaveFileName(
                    parent=self,
                    caption=f"Save recorded data ({size}Mb)",
                    directory=os.path.join(
                        self.workdir, f"colorevo_{now:%Y%m%d-%H%M%S}.h5"
                    ),
                    filter="HDF5 (*.h5)",
                )
                if fname:
                    try:
                        shutil.move(self._tmp_rec_file_name, fname)
                        msg = f"{size}Mb saved in {fname}"
                        self.statusBar().showMessage(msg)
                        break
                    except Exception as e:
                        logging.warning("Problem saving data: %r", e)

                # Data wasn't saved. Ask for confirmation about discarding
                btn = QtWidgets.QMessageBox.warning(
                    self,
                    "Discard last data?",
                    (
                        "Are you sure that you want to discard the "
                        + f"{size}Mb of data just recorded?\n"
                    ),
                    buttons=(
                        QtWidgets.QMessageBox.StandardButton.Save
                        | QtWidgets.QMessageBox.StandardButton.Discard
                    ),
                )
                if btn == QtWidgets.QMessageBox.StandardButton.Discard:
                    msg = f"Data in {self._tmp_rec_file_name} *not* saved"
                    self.statusBar().showMessage(msg)
                    break

    def addROI(self, classname, state=None, color=None):
        """Afactory of slightly customized ROIs"""

        if color is None:
            color = self.getNewColor()
        pen = pg.mkPen(color=color)

        if classname == "RectROI":
            if state is None:
                state = {"pos": [10, 10], "size": [200, 200], "angle": 0}
            roi = pg.RectROI(pos=[10, 10], size=[200, 200], pen=pen, removable=True)
            roi.addRotateHandle([1, 0], [0.5, 0.5])
        elif classname == "PolyLineROI":
            roi = pg.PolyLineROI(
                positions=[[0, 0], [100, 0], [0, 100]],
                closed=True,
                pen=pen,
                removable=True,
            )
        else:
            raise ValueError("Unsupported classname ({classname})")

        # connect the remove signal
        roi.sigRemoveRequested.connect(self.onROIRemoveRequested)
        # add the roi to the plot
        self.imgPlot.addItem(roi)
        # restore state if given
        if state is not None:
            roi.setState(state)
        # make sure the ROI is drawn on top
        roi.setZValue(10 + len(self.rois))
        # add it to the current rois list
        self.rois.append(roi)
        # also create the associated evo curves
        self.addEvoCurves(roi.pen)

    def addRectROI(self):
        state = None
        for previous in reversed(self.rois):
            if isinstance(previous, pg.RectROI):
                state = previous.saveState()
                break
        self.addROI("RectROI", state=state)

    def addPolyLineROI(self):
        state = None
        for previous in reversed(self.rois):
            if isinstance(previous, pg.PolyLineROI):
                state = previous.saveState()
                break
        self.addROI("PolyLineROI", state=state)

    def getNewColor(self):
        """
        Return a new QColor ensuring that it is not already used by any roi
        """
        existing_colors = [r.pen.color().getRgb() for r in self.rois]
        color = QtGui.QColor.fromHsvF(*next(self.itercolors))
        while color.getRgb() in existing_colors:
            color = QtGui.QColor.fromHsvF(*next(self.itercolors))
        return color

    def onROIRemoveRequested(self):
        roi = self.sender()
        idx = self.rois.index(roi)
        self.removeROI(idx)

    def removeROI(self, idx=-1):
        roi = self.rois.pop(idx)
        self.imgPlot.removeItem(roi)
        h_curve = self.h_curves.pop(idx)
        self.evoPlot_h.removeItem(h_curve)
        s_curve = self.s_curves.pop(idx)
        self.evoPlot_s.removeItem(s_curve)
        v_curve = self.v_curves.pop(idx)
        self.evoPlot_v.removeItem(v_curve)

    def removeAllRois(self):
        """remove all existing rois"""
        for _ in range(len(self.rois)):
            self.removeROI(-1)
        # restart the color generator
        self.itercolors = hsv_generator()

    def getROIsConfig(self):
        ret = []
        for roi in self.rois:
            cname = roi.__class__.__name__
            state = roi.saveState()
            color = roi.pen.color().getRgb()
            ret.append((cname, state, color))
        return ret

    def onSaveAllRoisAction(self):
        cfg = self.getROIsConfig()

        fname, flt = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            caption=f"Save {len(cfg)} ROIs",
            directory=self.workdir,
            filter="ROI template (*.roi)",
        )
        if fname:
            # add extension if it was not set
            if not fname.endswith(".roi"):
                fname += ".roi"
            with open(fname, "w") as f:
                json.dump(cfg, f)

    def restoreROIs(self, config):
        """
        Restore the ROIs defined in config. Config can be generated by
        `getROIsConfig()`
        """
        self.removeAllRois()

        # create new rois from config
        for c in config:
            self.addROI(*c)

    def onRestoreRoisAction(self):
        fname, flt = QtWidgets.QFileDialog.getOpenFileName(
            parent=self,
            caption="Load ROIs from file",
            directory=self.workdir,
            filter="ROI template (*.roi)",
        )
        if fname:
            with open(fname, "r") as f:
                self.restoreROIs(json.load(f))

    def addEvoCurves(self, pen):
        h_curve = pg.ScatterPlotItem(symbol="o", pen=pen, brush=pen.color(), size=7)
        s_curve = pg.ScatterPlotItem(symbol="o", pen=pen, brush=pen.color(), size=7)
        v_curve = pg.ScatterPlotItem(symbol="o", pen=pen, brush=pen.color(), size=7)
        self.evoPlot_h.addItem(h_curve)
        self.evoPlot_s.addItem(s_curve)
        self.evoPlot_v.addItem(v_curve)
        self.h_curves.append(h_curve)
        self.s_curves.append(s_curve)
        self.v_curves.append(v_curve)

    def onRefresh(self):
        # get last frame and time from the acq loop
        try:
            frame, t = self.acq_loop.data
        except Exception:
            return
        if t == self._last_processed:
            print(f"no new data since t={self._last_processed}. Skipping")
            return
        if frame is None:
            return
        self._last_processed = t
        # update the img and the plot
        self.img.setImage(bgr2rgb(frame))
        self.updateEvoPlot(frame, t)
        # report file size usage
        if self.recordAction.isChecked():
            size = os.path.getsize(self._tmp_rec_file_name) // 1_048_576  # in Mb
            msg = f"{self._tmp_rec_file_name}: {size}Mb"
            self.statusBar().showMessage(msg)

    def updateEvoPlot(self, frame, t):
        for roi, h_curve, s_curve, v_curve in zip(
            self.rois, self.h_curves, self.s_curves, self.v_curves
        ):
            selected = roi.getArrayRegion(frame, self.img)
            h, s, v = bgr2hsv(selected.astype(numpy.uint8)).mean(axis=(0, 1))
            h_curve.addPoints(x=(t,), y=(h,))
            s_curve.addPoints(x=(t,), y=(s,))
            v_curve.addPoints(x=(t,), y=(v,))

    def resetPlots(self):
        logging.info("Resetting plots")

        # black frame
        if self.img.image is not None:
            self.img.setImage(0 * self.img.image)

        # remove current color curves
        for h_curve in self.h_curves:
            self.evoPlot_h.removeItem(h_curve)
        self.h_curves = []
        for s_curve in self.s_curves:
            self.evoPlot_s.removeItem(s_curve)
        self.s_curves = []
        for v_curve in self.v_curves:
            self.evoPlot_v.removeItem(v_curve)
        self.v_curves = []

        # create new evo curves
        for roi in self.rois:
            self.addEvoCurves(roi.pen)

    def refreshEvoPlot(self, checked):
        if not checked:
            return

        self.resetPlots()

        if self._h5file is not None:
            self.onMainTimeSelectorChanged()  # re-show frame from cursor time

            msg = "Recalculating... "
            app = QtWidgets.QApplication.instance()

            frames = self._h5file["frames"]
            for i, (frame, t) in enumerate(zip(frames, self._h5times)):
                if not self.refreshEvoAction.isChecked():
                    self.statusBar().showMessage(msg + "Aborted", 3000)
                    break  # abort processing if refresh button is unchecked
                self.updateEvoPlot(frame, t)
                self.img.setImage(bgr2rgb(frame))
                self.statusBar().showMessage(
                    msg + f"({i+1}/{self._h5times.size}) frames", 3000
                )
                app.processEvents()

        # uncheck the button when finished
        self.refreshEvoAction.setChecked(False)

    def onMainTimeSelectorChanged(self):
        # find the index of the nearest frame to the selector
        xpos = self.time_selector_h.pos().x()
        with QtCore.QSignalBlocker(self.time_selector_s):
            self.time_selector_s.setPos(xpos)
        with QtCore.QSignalBlocker(self.time_selector_v):
            self.time_selector_v.setPos(xpos)
        adiff = numpy.abs(self._h5times - xpos)
        idx = numpy.argmin(adiff)
        # show frame at the found index
        frame = self._h5file["frames"][idx]
        self.img.setImage(bgr2rgb(frame))

    def onSlaveTimeSelectorChanged(self):
        self.time_selector_h.setPos(self.sender().pos().x())

    def setAcqFps(self, fps):
        if self.acq_loop is not None:
            self.acq_loop.period = 1 / fps

    @QtCore.pyqtSlot()
    def _setTmpRecFile(self, fname=None):
        """
        Sets the path and name of the file where the temporary data
        is recorded. If None given, a dialog is shown to prompt the user
        """
        if fname is None:
            fname, flt = QtWidgets.QFileDialog.getSaveFileName(
                parent=self,
                caption="File name for temporary data recording",
                directory=self._tmp_rec_file_name,
                filter="HDF5 (*.h5)",
            )
            if not fname:
                return
        self._tmp_rec_file_name = fname

    def _getTmpRecFile(self):
        return self._tmp_rec_file_name

    def saveSettings(self):
        """
        saves the application settings (so that they can be restored with
        :meth:`loadSettings`)
        """
        qsettings = QtCore.QSettings("colorevo", "colorevo")
        # main window geometry & state
        qsettings.setValue("geometry", self.saveGeometry())
        qsettings.setValue("state", self.saveState())
        # store the config dict
        qsettings.setValue("config", self.createQConfig())
        logging.info('Settings saved in "%s"', qsettings.fileName())

    def loadSettings(self):
        """
        restores the application settings previously saved with
        :meth:`saveSettings`.
        """
        qsettings = QtCore.QSettings("colorevo", "colorevo")

        # restore the app config
        try:
            self.applyQConfig(qsettings.value("config") or QtCore.QByteArray())
        except Exception as e:
            msg = (
                'Problem loading configuration from "%s".'
                + " Some settings may not be restored.\n Details: %r"
            )
            logging.warning(msg, qsettings.fileName(), e)
        self.restoreGeometry(qsettings.value("geometry") or QtCore.QByteArray())
        self.restoreState(qsettings.value("state") or QtCore.QByteArray())
        logging.info('Settings restored from "%s"', qsettings.fileName())

    def closeEvent(self, event):
        """
        Reimplemented from QMainWindow to clean and save settings on close
        """
        self.saveSettings()

        if self.acq_loop is not None:
            self.acq_loop.stop()
            self.acq_loop.cap.release()
        if self.acq_thread is not None:
            logging.info("Shutting down acquisition loop...")
            self.acq_thread.join()
        if self._h5file is not None:
            self._h5file.close()
        QtWidgets.QMainWindow.closeEvent(self, event)


@click.command()
@click.version_option(version=__version__)
@click.option(
    "--keep-alive/--no-keep-alive",
    default=False,
    help="do not close GUI on exceptions",
)
@click.option("--reset", is_flag=True, help="reset configuration to default settings")
def main(keep_alive, reset):
    # Avoid closing GUI on exception if requested
    if keep_alive:
        import traceback

        def excepthook(etype, value, tb):
            traceback.print_exception(etype, value, tb)

        sys.excepthook = excepthook

    app = QtWidgets.QApplication(sys.argv)
    w = ColorEvoMain(load_settings=not reset)
    w.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()
